/*

	m - command line mailer for BeOS, supporting plug-in editors,
	composition, and reading of mail while being multi-threaded and
	all that sort of thing.
	
	This program is completely free software: you can redistribute it
	and/or modify it under with no conditions. In addition, IT IS 
	FREE OF ANY SORT OF WARRANTY WHATSOEVER, NEITHER EXPRESSED NOR
	IMPLIED. This clause should not be necessary, but there may 
	actually be bugs in this software, in which case the former
	statement can be employed (ie, you can fix it yourself).
	
	Of course, if you would like, correspondence can be made for the
	time being to , 5114 Ellicott Hall, College Park, MD 20742-9351.
	Perhaps if we have time, we will improve this and add a GUI feature
	and such.
	
	user24601@veritas.dorm.umd.edu
*/

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <NetKit.h>
#include <StorageKit.h>
#include <Be.h>
#include <AppKit.h>

//----------------------------------------------------------------
//----------------------------------------------------------------
//Class mapp - the real application class
//----------------------------------------------------------------
//----------------------------------------------------------------

#define	C_QUITS			"quit"
#define C_QUITC			"q"
#define C_EXITS			"exit"
#define C_EXITC			"x"
#define C_HELPS			"help"
#define C_HELPC			"?"
#define C_SWITCHS		"switch"
#define C_SWITCHC		"s"
#define C_DELS			"delete"
#define C_DELC			"d"
#define C_PURGES		"purge"
#define C_PURGEC		"p"
#define C_CHECKS		"check"
#define C_CHECKC		"c"
#define C_LISTS			"list"
#define C_LISTC			"l"
#define C_ABOUTS		"about"
#define C_ABOUTC		"a"
#define C_HOSTS			"host"
#define C_HOSTC			"h"
#define C_PURGE_BOX		3001
#define C_RESTORE_BOX	3002
#define MAX_LINE 		B_MAX_HOST_NAME_LENGTH
#define MAX_HEADERS		12


class mapp : public BApplication {

public:
					mapp();
	virtual			~mapp();
	void			ArgvReceived(int argc, char** argv);
	void			MessageReceived(BMessage* recMessage);
	void			ReadyToRun();
	void			readPrefs();

	void			newMbox();
	void			switchMbox();
	void			displayMbox();
	void			engine();
	void			getMessage(int);
	void			printFields(BMailMessage* 	theMail, 
								const char*		theField, 
								int				theLength); 
	void			cleanBox(int cleanType);
	void			setDeleted(int num);
	void			changeHost();
	void			displayMessage(int num);
	void			showHelp();
	void			showAbout();
	void			checkMail();
	int				doQuitRoutine();
	void			doUpdateQuery();
	void			doPurgeQuery();

	BDatabase		*mdb;
	BTable			*mtable;
	BQuery			*currMbox;
	int				currNum;
	BMailMessage	*currMail;

	char			userbox[B_MAX_USER_NAME_LENGTH];


private:
	char			toField[B_MAX_USER_NAME_LENGTH];
	mail_account	omacc;
	mail_account	tmpmacc;
	struct fdisplay{
		char 	field[B_MAX_USER_NAME_LENGTH];
		int		length;
	};
	fdisplay		mlabels[3], mdisplay[MAX_HEADERS];
	int				mdisnum;
	long 			*messnum;
	bool			delMod, readMod;
	int				error;
		
	enum acttype {
		O_READ,
		O_SEND
	} 				action;
	enum mmode {
		O_IB,
		O_MB,
		O_ALL,
		O_USER,
		O_ERR,
		O_PEND,
		O_DEL,
		O_USED
	} 				mbtype;
};

//----------------------------------------------------------------
//main
//----------------------------------------------------------------

int main() {

	mapp *m = new mapp();
	
	m->Run();
	delete m;
	return 1;
}
	
//----------------------------------------------------------------
//Constructor and Destructor
//----------------------------------------------------------------

mapp::mapp() : BApplication('m') {

	
	mdb		= boot_volume().Database();
	mtable	= mdb->FindTable("E-Mail");
	mbtype	= O_IB;
	action	= O_READ;
	currMbox = NULL;
}

mapp::~mapp() {

	if(currMbox) {
		delete currMbox;
		currMbox = NULL;
	}
	set_mail_account(&omacc);
}

//----------------------------------------------------------------
//MessageReceived - to catch the query update messages
//----------------------------------------------------------------

void mapp::MessageReceived(BMessage* recMessage) {

	switch(recMessage->what) {
		case B_RECORD_ADDED    :
		case B_RECORD_REMOVED  :
		case B_RECORD_MODIFIED :
			update_query(recMessage);
			break;
		default:
			BApplication::MessageReceived(recMessage);
			break;
	}
}

//----------------------------------------------------------------
//ArgvReceived - process command line arguments
//----------------------------------------------------------------

void mapp::ArgvReceived(int argc, char** argv) {

	if(argc!=0) {
		fprintf(stderr, "Usage: m\n");
		error = 1;
	}
	else error = 0;
}

//----------------------------------------------------------------
//ReadyToRun
//----------------------------------------------------------------

void mapp::ReadyToRun() {

	if(!error) {
		get_mail_account(&omacc);
		tmpmacc = omacc;
		checkMail();
		newMbox();
		printf("\nYou have %d messages in the InBox\n\n",
					currMbox->CountRecordIDs());
		readPrefs();
		engine();
	}
	PostMessage(B_QUIT_REQUESTED);
}

//----------------------------------------------------------------
//readPrefs - reads settings from a preferences file
//----------------------------------------------------------------

void mapp::readPrefs() {

	FILE *prefs = fopen("/boot/home/.mprefs", "r");
	char inC;
	while((inC=getc(prefs))!='^');
	for(int i=0; i<3; i++) {
		int j=0;
		while((inC=getc(prefs))!='^' && j<B_MAX_USER_NAME_LENGTH) {
			mlabels[i].field[j]=inC;
			j++;
		}
		mlabels[i].field[j-1]='\0';
		while(inC!='^') inC=getc(prefs);
		fscanf(prefs, "%d", &mlabels[i].length);
		while((inC=getc(prefs))!='^');
	}
	while((inC=getc(prefs))!='^');
	int i=0;
	while((inC)!='#' && i<MAX_HEADERS) {
		int j=0;
		while((inC=getc(prefs))!='^' && inC!='#' && j<B_MAX_USER_NAME_LENGTH) {
			mdisplay[i].field[j]=inC;
			j++;
		}
		mdisplay[i].field[j-1]='\0';
		while(inC!='^' && inC!='#') inC=getc(prefs);
		mdisplay[i].length = 0;
		i++;
	}
	mdisnum = i;
	fclose(prefs);
}

//----------------------------------------------------------------
//checkMail - checks for mail
//----------------------------------------------------------------

void mapp::checkMail() {

	printf("Checking for mail\n");
	set_mail_account(&tmpmacc);
	check_for_mail(messnum);
}
		

//----------------------------------------------------------------
//newMbox - generates new mailbox according to current mbtype
//----------------------------------------------------------------

void mapp::newMbox() {

	if(currMbox) delete currMbox;
	currMbox = new BQuery();

	switch(mbtype) {
	
		case O_IB:
			currMbox->AddTree(mtable);
			currMbox->PushField("Status");
			currMbox->PushString("*New*");
			currMbox->PushOp(B_EQ);
			currMbox->PushField("Status");
			currMbox->PushString("*Used*");
			currMbox->PushOp(B_EQ);
			currMbox->PushOp(B_OR);
			currMbox->PushField("Status");
			currMbox->PushString("*Del*");
			currMbox->PushOp(B_EQ);
			currMbox->PushOp(B_OR);
			break;
		case O_MB:
			currMbox->AddTree(mtable);
			currMbox->PushField("Status");
			currMbox->PushString("*New*");
			currMbox->PushOp(B_EQ);
			currMbox->PushField("Status");
			currMbox->PushString("*Read*");
			currMbox->PushOp(B_EQ);
			currMbox->PushOp(B_OR);
			currMbox->PushField("Status");
			currMbox->PushString("*Del*");
			currMbox->PushOp(B_EQ);
			currMbox->PushOp(B_OR);
			break;
		case O_ALL:
			currMbox->AddTree(mtable);
			currMbox->PushOp(B_ALL);
			break;
		case O_USER:
			currMbox->AddTree(mtable);
			currMbox->PushField("From");
			currMbox->PushString(userbox);
			currMbox->PushOp(B_EQ);
			break;
		case O_ERR:
			currMbox->AddTree(mtable);
			currMbox->PushField("Status");
			currMbox->PushString("*Error*");
			currMbox->PushOp(B_EQ);
			break;
		case O_PEND:
			currMbox->AddTree(mtable);
			currMbox->PushField("Status");
			currMbox->PushString("*Pending*");
			currMbox->PushOp(B_EQ);
			currMbox->PushField("Status");
			currMbox->PushString("*Sent*");
			currMbox->PushOp(B_EQ);
			currMbox->PushOp(B_OR);
			currMbox->PushField("Status");
			currMbox->PushString("*Del*");
			currMbox->PushOp(B_EQ);
			currMbox->PushOp(B_OR);
			break;
		case O_DEL:
			currMbox->AddTree(mtable);
			currMbox->PushField("Status");
			currMbox->PushString("*Del*");
			currMbox->PushOp(B_EQ);
			break;
		case O_USED:
			currMbox->AddTree(mtable);
			currMbox->PushField("Status");
			currMbox->PushString("*Used*");
			currMbox->PushOp(B_EQ);
			break;
	}
	
	currMbox->Fetch();
}

//----------------------------------------------------------------
//switchMbox - reads in choice of box, and calls newMbox
//----------------------------------------------------------------

void mapp::switchMbox() {

	char 	inString[B_MAX_USER_NAME_LENGTH];
	mmode 	tempMBT = O_USER;
	int 	i;
	char	inC;
	
	printf("!, >, <, *, @, or a string [ return exits ]--> ");
	fflush(stdout);
	i = 0;
	while(i<B_MAX_USER_NAME_LENGTH && (inC=getchar())!='\n') {
		inString[i] = inC;
		i++;
	}
	inString[i]='\0';
	fflush(stdin);
	switch(inString[0]) {
	
		case '!':
			tempMBT = O_IB;
			break;
		case '>':
			tempMBT = O_MB;
			break;
		case '*':
			tempMBT = O_ALL;
			break;
		case '@':
			tempMBT = O_ERR;
			break;
		case '<':
			tempMBT = O_PEND;
			break;
		case '\0':
			tempMBT = mbtype;
			break;
		default:
			userbox[0] = '*';
			i =0;
			while(i<B_MAX_USER_NAME_LENGTH-2 && inString[i]!='\0') {
				userbox[i+1] = inString[i];
				i++;
			}
			userbox[i++] ='*';
			userbox[i+1]   ='\0';
			doPurgeQuery();
			doUpdateQuery();
			mbtype = O_USER;
			newMbox();
			delMod = FALSE;
			readMod = FALSE;
			break;
	}
	if(tempMBT!=mbtype) {
		doPurgeQuery();
		doUpdateQuery();
		mbtype = tempMBT;
		newMbox();
		delMod = FALSE;
		readMod = FALSE;
	}
}

//----------------------------------------------------------------
//changeHost - change to a different mail host
//----------------------------------------------------------------

void mapp::changeHost() {

	char inLine[MAX_LINE];
	char inC;
	int i;

	printf("\nEnter new host information [ return keeps current value ]");
	printf("\nusername [ %s ]--> ", tmpmacc.pop_name);
	fflush(stdout);
	i = 0;
	while(i<MAX_LINE-1 && (inLine[i]=getchar())!='\n')
		i++;
	inLine[i]='\0';
	fflush(stdin);
	if(inLine[0]!='\0')
		for(int i=0;i<MAX_LINE,(tmpmacc.pop_name[i]=inLine[i])!='\0';i++);

	printf("\npassword [ <hidden> ]--> ");
	fflush(stdout);
	i = 0;
	while(i<MAX_LINE-1 && (inLine[i]=getchar())!='\n')
		i++;
	inLine[i]='\0';
	fflush(stdin);
	if(inLine[0]!='\0')
		for(i=0;i<MAX_LINE,(tmpmacc.pop_password[i]=inLine[i])!='\0';i++);

	printf("\npop host [ %s ]--> ", tmpmacc.pop_host);
	fflush(stdout);
	i = 0;
	while(i<MAX_LINE-1 && (inLine[i]=getchar())!='\n')
		i++;
	inLine[i]='\0';
	fflush(stdin);
	if(inLine[0]!='\0')
		for(int i=0;i<MAX_LINE,(tmpmacc.pop_host[i]=inLine[i])!='\0';i++);

	printf("\nsmtp host [ %s ]--> ", tmpmacc.smtp_host);
	fflush(stdout);
	i = 0;
	while(i<MAX_LINE-1 && (inLine[i]=getchar())!='\n')
		i++;
	inLine[i]='\0';
	fflush(stdin);
	if(inLine[0]!='\0')
		for(int i=0;i<MAX_LINE,(tmpmacc.smtp_host[i]=inLine[i])!='\0';i++);
}

//----------------------------------------------------------------
//getMessage - finds the message of the specified index
//----------------------------------------------------------------

void mapp::getMessage(int num) {

	record_id 	rec_id;
	BRecord		*tempRec;

	if(mtable) {
		if(currMail) {
			delete currMail;
			currMail = NULL;
		}
		rec_id   = currMbox->RecordIDAt(num-1);
		tempRec  = new BRecord(mdb, rec_id);
		currMail = new BMailMessage(tempRec);
		delete tempRec;
	}
}

//----------------------------------------------------------------
//deleteMessage - sets message of specified index to be deleted
//----------------------------------------------------------------

void mapp::setDeleted(int num) {

	record_id		rec_id;
	BRecord			*tmpRec;
	rec_id	= currMbox->RecordIDAt(num-1);
	tmpRec	= new BRecord(mdb, rec_id);
	tmpRec->SetString("Status", "Del");
	tmpRec->Commit();
	delete tmpRec;
	delMod = TRUE;
}

//----------------------------------------------------------------
//cleanBox - cleans up the mailbox
//----------------------------------------------------------------

void mapp::cleanBox(int cleanType) {

	BRecord		*tmpRec;
	int			i=0;
	int			num_recs;
	mmode		tmpmbtype = mbtype;
	record_id	rec_id;
	
	if(mtable) {
		if(currMail) {
			delete currMail;
			currMail = NULL;
		}
	}
	mbtype = O_DEL;
	newMbox();
	num_recs = currMbox->CountRecordIDs();
	while(i<num_recs) {
		rec_id = currMbox->RecordIDAt(i);
		tmpRec	= new BRecord(mdb, rec_id);
		if(cleanType==C_PURGE_BOX)
			tmpRec->Remove();
		else if(cleanType==C_RESTORE_BOX) {
			tmpRec->SetString("Status", "Used");
			tmpRec->Commit();
		}
		i++;
		delete tmpRec;
	}
	mbtype = tmpmbtype;
	newMbox();
	delMod = FALSE;
}

//----------------------------------------------------------------
//displayMbox - numerically indexes the current mailbox
//----------------------------------------------------------------

void mapp::displayMbox() {

	int 		num;
	BRecord 	*tmpRec;
	
	currMbox->Fetch();
	printf("\n     ");
	for(int i=0;i<3;i++)
		printf("%-*s  ", mlabels[i].length, mlabels[i].field);
	putchar('\n');
	for(int i=1;i<=80;i++) printf("-");
	putchar('\n');
	int num1 = currMbox->CountRecordIDs();
	for(int i=1;i<=num1;i++) {
		getMessage(i);
		if(currMail) {
			tmpRec=new BRecord(mdb, currMbox->RecordIDAt(i-1));
			if(!strcmp(tmpRec->FindString("Status"), "Del"))
				putchar('D');
			else
				putchar(' ');
			printf("%3d ", i);
			for(int j=0;j<=2;j++) {
				printFields(currMail, mlabels[j].field, mlabels[j].length);
				printf("  ");
			}
			putchar('\n');
		}
		else {
			printf("Oops. Something seems to have gone wrong with message %d.\n",i);
		}
	}
	if(num1==0)
		printf("\n       This box is empty\n");
}

//----------------------------------------------------------------
//displayMessage - prints out the message
//----------------------------------------------------------------

void mapp::displayMessage(int num) {

	getMessage(num);
	for(int i=0;i<mdisnum;i++) {
		printf("%s", mdisplay[i].field);
		printFields(currMail, mdisplay[i].field, mdisplay[i].length);
		putchar('\n');
	}
	putchar('\n');
	
	char *content;
	long length;
	currMail->FindField((char *) B_MAIL_CONTENT, &content, &length);
	for(int i=0;i<length;i++)
		putchar(content[i]);
	putchar('\n');
	
	record_id		rec_id;
	BRecord			*tmpRec;
	rec_id	= currMbox->RecordIDAt(num-1);
	tmpRec	= new BRecord(mdb, rec_id);
	if(strcmp(tmpRec->FindString("Status"), "New")!=0) {
		tmpRec->SetString("Status", "Used");
		tmpRec->Commit();
	}
	readMod = TRUE;
	delete tmpRec;
}

//----------------------------------------------------------------
//printFields - nicely formatted output of certain message fields
//----------------------------------------------------------------

void mapp::printFields(BMailMessage* theMail,
					   const char*   theField,
					   int			 theLength) {
	long retLength;
	char *retData;
	char *outString;
	long outLength;
	if(theMail->FindField((char *) theField, &retData, &retLength)!=B_NO_ERROR) {
		outString = "***";
		outLength = strlen(outString);
	}
	else {
		outString = (char *) retData;
		outLength = retLength;
	}
	if(theLength!=0) {
		if(outLength > theLength) {
			outLength = theLength;
		}
	}

	for(int i=0;i<outLength;i++) putchar(outString[i]);
	if(theLength!=0) 
		for(int i=outLength;i<theLength;i++) putchar(' ');
}
				
//----------------------------------------------------------------
//doPurgeQuery - delete the marked messages?
//----------------------------------------------------------------

void mapp::doPurgeQuery() {

	char inC;
	if(delMod) {
		printf("\nShall we remove the marked messages? ([y] / n)--> ");
		fflush(stdout);
		inC = getchar();
		fflush(stdin);
		if(!(inC=='n' || inC=='N'))
			cleanBox(C_PURGE_BOX);
		else
			cleanBox(C_RESTORE_BOX);
	}
}

//----------------------------------------------------------------
//doUpdateQuery - 'Used' messages are converted to 'Read'
//----------------------------------------------------------------

void mapp::doUpdateQuery() {

	int 		i = 0;
	int			num_recs;
	record_id	rec_id;
	BRecord		*tmpRec;
	mmode		tmpmbtype = mbtype;
	
	if(readMod) {
	    mbtype = O_USED;
	    newMbox();
	    num_recs = currMbox->CountRecordIDs();
	    while(i<num_recs) {
	        rec_id  = currMbox->RecordIDAt(i);
	        tmpRec  = new BRecord(mdb, rec_id);
	        tmpRec->SetString("Status", "Read");
	        tmpRec->Commit();
	        i++;
	        delete tmpRec;
	    }
	    mbtype = tmpmbtype;
	    newMbox();
	}
}

//----------------------------------------------------------------
//doQuitRoutine - Quitting and stuff
//----------------------------------------------------------------

int mapp::doQuitRoutine() {

	char inC;
	
	printf("\nAre you sure you want to quit? ([y] / n)--> ");
	fflush(stdout);
	inC = getchar();
	fflush(stdin);
	if(!(inC=='n' || inC=='N')) {
		doPurgeQuery();
		doUpdateQuery();
		return 1;
	}
	return 0;
}

//----------------------------------------------------------------
//engine - this is the general interface to the program
//----------------------------------------------------------------

void mapp::engine() {
	
	char cmd[MAX_LINE];
	char c;
	int  i;
	int fini = 0;
	int num;

	while(!fini) {	
		printf("\n--> ");
		fflush(stdout);
		fflush(stdin);
		i = 0;
		while(i<MAX_LINE-1 && (c=getchar())!='\n') {
			cmd[i]=tolower(c);
			i++;
		}
		fflush(stdin);
		cmd[i]='\0';
		if(i!='\0') {
			if(!strcmp(cmd, C_HELPS) || !strcmp(cmd, C_HELPC))
				showHelp();
			else if(!strcmp(cmd, C_ABOUTS) || !strcmp(cmd, C_ABOUTC)) 
				showAbout();
			else if(!strcmp(cmd, C_QUITS) || !strcmp(cmd, C_QUITC)
					|| !strcmp(cmd, C_EXITS) || !strcmp(cmd, C_EXITC))
				fini = doQuitRoutine();
			else if(!strcmp(cmd, C_CHECKS) || !strcmp(cmd, C_CHECKC))
				checkMail();
			else if(!strcmp(cmd, C_LISTS) || !strcmp(cmd, C_LISTC))
				displayMbox();
			else if(!strcmp(cmd, C_PURGES) || !strcmp(cmd, C_PURGEC)) {
				cleanBox(C_PURGE_BOX);
				doUpdateQuery();
			}
			else if(!strcmp(cmd, C_DELS) || !strcmp(cmd, C_DELC)) {
				num = 0;
				printf("\nMessage Num? (return escapes)--> ");
				fflush(stdout);
				char tmpnum[10];
				i =0;
				while((c=getchar())!='\n')
					if(isdigit(c))
						tmpnum[i++] = c;
				tmpnum[i] = '\0';
				if(tmpnum[0]!='\0') {
					num = atoi(tmpnum);
					if(num<=currMbox->CountRecordIDs()) {
						setDeleted(num);
						displayMbox();
					}
					else
						printf("\nHmm...there does not seem to be a message with that number.\n");
				}
			}
			else if(!strcmp(cmd, C_SWITCHS) || !strcmp(cmd, C_SWITCHC))
				switchMbox();
			else if(!strcmp(cmd, C_HOSTS) || !strcmp(cmd, C_HOSTC)) {
				changeHost();
			}
			else {
				int j = 0;
				while(isdigit(cmd[j++]) && j<=i);
				if(j>i) {
					num = atoi(cmd);
					if(num<=currMbox->CountRecordIDs()) 
						displayMessage(num);
					else
						printf("\nHmm...there does not seem to be a message with that number.\n");
				}
				else printf("Umm...please type 'help' for assistance");
			}
		}
	}
}
//----------------------------------------------------------------
//showHelp and showAbout - this should be evident
//----------------------------------------------------------------

void mapp::showHelp() {

	printf("\n%s\n", "        m mailer commands\n" \
		" help, ?      This stuff\n" \
		" about, a     Legal nonsense\n" \
		" check, c     Get new mail from current server\n" \
		" list, l      List current mailbox\n" \
		" ###          Display message ###\n" \
		" delete, d    Delete a message\n" \
		" purge, p     Remove the deleted messages\n" \
		" host, h      Change mailhost\n\n" \
		" switch, s    Switch mailbox (different query...)\n" \
		"       ! - Inbox             > - Received\n" \
		"       < - Sent              * - All\n" \
		"       @ - Error          abcd - Username contains abcd\n");
}

void mapp::showAbout() {

	printf("\n%s\n", " About - Revision 0.82a\n" \
		"	m - command line mailer for BeOS, supporting plug-in editors,\n" \
		" composition, and reading of mail while being multi-threaded and\n" \
		" all that sort of thing.\n" \
		"\n" \
		" This program is completely free software: you can redistribute it\n" \
		" and/or modify it under with no conditions. In addition, it is\n" \
		" FREE OF ANY SORT OF WARRANTY WHATSOEVER, NEITHER EXPRESSED NOR\n" \
		" IMPLIED. This clause should not be necessary, but there may\n" \
		" actually be bugs in this software, in which case the former\n" \
		" statement can be employed (ie, you can fix it yourself).\n" \
		"\n" \
		" Of course, if you would like, correspondence can be made for the\n" \
		" time being to , 5114 Ellicott Hall, College Park, MD 20742-9351.\n" \
		" Perhaps if we have time, we will improve this and add a GUI feature\n" \
		" and such.\n");
}
